## Automatically adapted for numpy.oldnumeric Feb 04, 2011 by 

# Name: xdp/interpolate.py
# Purpose: Provides linear interpolation
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
Linear interpolation.
"""


import sys
import bisect
import numpy.oldnumeric as Numeric


__all__ = ['interpolate', 'lint']


def interpolate(x, xVec, yVec):
    """
    Interpolate one dependent value.

    @param x: independent value to interpolate on
    @type  x: C{int} or C{float}
    @param xVec: independent values
    @type  xVec: C{Numeric} vector
    @param yVec: dependent values
    @type  yVec: C{Numeric} vector
    @return: dependent value at C{x}
    @rtype:  C{int} or C{float}
    """
    # nasty hack to support using lint() with fit.py's lsfit()
    if hasattr(x, 'value'):
        xDerivVar = x
        x = x.value
    else:
        xDerivVar = x

    b = bisect.bisect(xVec, x)
    if b == 0:
        return yVec[0]
    elif b == len(xVec):
        return yVec[-1]
    else:
        b = b - 1

    x1 = xVec[b]
    y1 = yVec[b]
    x2 = xVec[b+1]
    y2 = yVec[b+1]

    m = float(y2 - y1)/float(x2 - x1)
    return m*xDerivVar + (y1 - m*x1)


def lint(xVec, yVec, xRange):
    """
    Perform a linear interpolation.

    If C{yVec} is a matrix, interpolation will be performed for each column of
    the matrix.

    @param xVec: independent values
    @type  xVec: C{Numeric} vector
    @param yVec: dependent values
    @type  yVec: C{Numeric} vector or matrix
    @param xRange: abscissa
    @type  xRange: C{Numeric} vector
    @return: interpolated dependent values
    @rtype:  C{Numeric} vector
    """

    xVecLength = len(xVec)
    xRangeLength = len(xRange)
    newY = Numeric.zeros( (xRangeLength,), 'd')

    for i in range(0, xRangeLength):
        x = xRange[i]
        b = bisect.bisect(xVec, x)
        if b == 0:
            newY[i] = yVec[0]
            continue
        elif b == xVecLength:
            newY[i] = yVec[-1]
            continue
        else:
            b = b - 1

        x1 = xVec[b]
        y1 = yVec[b]
        x2 = xVec[b+1]
        y2 = yVec[b+1]

        m = float(y2 - y1)/float(x2 - x1)
        newY[i] = m*x + (y1 - m*x1)
    return newY
