## Automatically adapted for numpy.oldnumeric Feb 04, 2011 by 

# Name: xdp/interpolate/stat_linear.py
# Purpose: Provides linear interpolation with statistics
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
Linear interpolation that preserves the dependant variable's relative error.
"""


import sys
import math
import bisect
import numpy.oldnumeric as Numeric


__all__ = ['interpolate', 'stat_lint']


def interpolate(x, xVec, yVec, yErr):
    """
    Interpolate one dependent value, preserving its relative error.

    @param x: independent value to interpolate on
    @type  x: C{int} or C{float}
    @param xVec: independent values
    @type  xVec: C{Numeric} vector
    @param yVec: dependent values
    @type  yVec: C{Numeric} vector
    @param yErr: relative errors
    @type  yErr: C{Numeric} vector
    @return: dependent value and relative error at C{x}
    @rtype:  C{2-tuple} of the form C{(y, err)}
    """

    b = bisect.bisect(xVec, x)
    if b == 0:
        return (yVec[0], yErr[0])
    elif b == len(xVec):
        return (yVec[-1], yErr[-1])
    else:
        b = b - 1

    x1 = xVec[b]
    y1 = yVec[b]
    x2 = xVec[b+1]
    y2 = yVec[b+1]

    m = float(y2 - y1)/float(x2 - x1)
    y = m*x + (y1 - m*x1)

    y1err = yErr[b]
    y2err = yErr[b+1]

    a = ((x2 - x)/(x2 - x1))**2 * (y1err * y1)
    b = ((x - x1)/(x2 - x1))**2 * (y2err * y2)
    y_rel = math.sqrt(a - b)

    return (y, y_rel)


def stat_lint(xVec, yVec, xRange, yErr):
    """
    Perform a linear interpolation that preserves relative error.

    @param xVec: independent values
    @type  xVec: C{Numeric} vector
    @param yVec: dependent values
    @type  yVec: C{Numeric} vector
    @param xRange: abscissa
    @type  xRange: C{Numeric} vector
    @param yErr: relative errors
    @type  yErr: C{Numeric} vector
    @return: interpolated dependent values and their relative errors
    @rtype:  C{2-tuple} of C{Numeric} vectors
    """

    newY = Numeric.zeros( (len(xRange),), 'd')
    newYerr = Numeric.zeros( (len(xRange),), 'd')

    for i in range(0, len(xRange)):
        x = xRange[i]
        b = bisect.bisect(xVec, x)
        if b == 0:
            newY[i] = yVec[0]
            newYerr[i] = yErr[0]
            continue
        elif b == len(xVec):
            newY[i] = yVec[-1]
            newYerr[i] = yErr[-1]
            continue
        else:
            b = b - 1

        x1 = xVec[b]
        y1 = yVec[b]
        x2 = xVec[b+1]
        y2 = yVec[b+1]

        m = float(y2 - y1)/float(x2 - x1)
        newY[i] = m*x + (y1 - m*x1)

        y1err = yErr[b]
        y2err = yErr[b+1]

        a = ((x2 - x)/(x2 - x1))**2 * (y1err * y1)
        b = ((x - x1)/(x2 - x1))**2 * (y2err * y2)
        newYerr[i] = math.sqrt(abs(a - b))
    return (newY, newYerr)
