## Automatically adapted for numpy.oldnumeric Feb 04, 2011 by 

# Name: meansum.py
# Purpose: utilities for writing XDP applications which mean & sum data
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""
Write applications which sum and average datasets.
"""


import os
import numpy.oldnumeric as Numeric

import xdp
from xdp.io import readFile
from xdp.interpolate.linear import interpolate
from xdp.data import clean, checkRange, coljoin, trim
from xdp.transformation import Mapping, CleanTransform, LintTransform, TrimTransform


class MeansumError(Exception):
    """
    Exception raised to indicate that some meansum-specific error occured.
    """

    pass


class MeansumMapping:
    """
    Map experiment data onto independant and dependant values using Python
    expressions and C{Data.evaluate}.

    The destination columns are:
     - I{X} S{->} independent values used in regridding
     - I{Y} S{->} dependent values used in regridding

    @todo: make this is more general tool
    """

    def __init__(self, xExpr, yExpr):
        """
        Create a new MeansumMapping object.

        @param xExpr: Python expression for calculating the independant values
        @type  xExpr: C{str}ing
        @param yExpr: Python expression for calculating the dependant values
        @type  yExpr: C{str}ing
        """

        self.__xExpr = xExpr
        self.__yExpr = yExpr

    def __call__(self, data):
        """
        Apply this mapping to a dataset, creating a new dataset that contains
        the independent and dependent values.

        @param data: dataset
        @type  data: C{Data}
        @return: I{X} and I{Y} dataset
        @rtype:  C{Data}

        @raise EvaluationError: one of the Python expressions produced an error
        or did not result in a C{Numeric} vector
        """

        try:
             x =  data.evaluate(self.__xExpr)
        except xdp.EvaluationError, e:
            raise xdp.EvaluationError(
                '%s in independant variable\'s expression' % e)

        if isinstance(x, not Numeric.ArrayType) or len(x.shape) != 1:
            raise xdp.EvaluationError(
    'independant variable\'s expression is not a one-dimensional Numeric array')

        try:
            y = data.evaluate(self.__yExpr)
        except xdp.EvaluationError, e:
            raise xdp.EvaluationError(
                '%s in dependant variable\'s expression' % e)

        if isinstance(y, not Numeric.ArrayType) or len(y.shape) != 1:
            raise xdp.EvaluationError(
    'dependant variable\'s expression is not a one-dimensional Numeric array')

        result = xdp.Data(X=x, Y=y)
        return result


class ExperimentData(xdp.Data):
    """
    Experiment data loaded from a file for summing or averaging.

    @ivar _path: full path and name of dataset's source file
    @type _path: C{str}ing
    @ivar _name: name of dataset's source file
    @type _name: C{str}ing
    @ivar _shift: relative X-axis shift
    @type _shift: C{float} or C{None}
    @ivar header: header of this dataset's source file
    @type header: C{Sff}, C{Xafs}, X{Text}, or C{None}
    """

    def __init__(self, path='', mapping=None):
        """
        Create an empty ExperimentData object or make one by loading the
        experiment data from a file.

        @param path: file to load
        @type  path: C{str}ing
        @param mapping: mapping to apply to loaded dataset
        @type  mapping: C{Mapping}

        @note: If the file being loaded does not have I{X} and I{Y} columns,
        the appropriate mapping must be supplied.
        """

        xdp.Data.__init__(self)

        self._path = path
        self._shift = None
        self._header = None
        self._file = os.path.basename(path)

        if path:
            header, data = readFile(path)
            self._header = header
            clean = CleanTransform('X', 'Y')

            if mapping:
                data = clean(mapping(data))
            else:
                data = clean(data)

            for columnName in data.columns():
                self.add(columnName, data.get(columnName))

#XXX: continue documenting here

def shift(expData, start, end):
    """Perform energy shift correction relative to the first ExperimentData
    object of the sequence 'expData', updating the elements of that sequence to
    reflect the energy shifts.

    The arguments 'start' and 'end' specify the energy range over which to
    perform the shift correction.
    """
    Min = max([ min(x.X) for x in expData ])
    Max = min([ max(x.X) for x in expData ])

    try:
        start, end = checkRange(Min, Max, start, end)
    except ValueError, e:
        raise ValueError('I.V.S. %s' % e)

    trim = TrimTransform('X', 'Y', start, end)
    base = trim(expData[0])

    for exp in expData[1:]:
        other = trim(exp)
        params, error = findIVShift(base, other, None, None, 0.001, 1.0,
            0.001)

        exp._shift = params[0]
        exp.X += exp._shift


def sum(expData, grid, start, end):
    """Sum experiment data by regridding Y for each ExperimentData object in
    the 'expData' sequence to the one-dimensional Numeric array 'dep' using
    linear interpolation and then summing the results.

    The return value is an xdp.Data object containing an X column
    correspoding to 'grid' and the sum of the regridded Y columns of all of
    'expData'.
    """
    Min = max([ min(x.X) for x in expData ])
    Max = min([ max(x.X) for x in expData ])

    try:
        start, end = checkRange(Min, Max, start, end)
    except ValueError, e:
        raise MeansumError, ('regridding %s' % e)

    sum = xdp.Data()
    sum.add('X', grid)
    sum.add('Y', Numeric.zeros(grid.shape, 'd'))

    interpolate = LintTransform('X', 'Y', grid)

    for exp in expData:
        exp = interpolate(exp)
        sum.Y += exp.Y

    return sum


def mean(expData, grid, start, end):
    """Find the arithmetic mean of experiemnt data by summing 'expData'
    using sum() and then dividing the result's Y by the number of elements.

    The return value is an xdp.Data object containing an X column
    corresponding to 'grid' and the arithmetic mean of the regridded Y
    columns of all of 'expData'.
    """
    result = sum(expData, grid, start, end)
    result.Y /= float(len(expData))
    return result


#
# Least-squares fitting... requires Scientific Python
#

class IVShiftFunc:
    """Private utility class implementing the callback function for
    findIVShift().
    """
    def __init__(self, x, y):
        """Initialize a new instance of the callback function.  'x' and
        'y' are the vectors of the second data set from a call to
        findIVShift().
        """
        self.__x = x
        self.__y = y

    def __call__(self, params, energy):
        """Implements the callback function for the nonlinear least squared fit.
        """
        dX = params[0]
        amplitude = params[1]
        dY = params[2]
        y = interpolate(energy + dX, self.__x, self.__y)
        return (y * amplitude) + dY


def findIVShift(base, other, start, end, dX, amplitude, dY):
    """Finds the dX, amplitude, and dY that fits two data sets, 'base' and
    'other'.

    The xdp.Data object 'other' is fit the xdp.Data object 'base' using the
    columns 'X' and 'Y'.

    'start' and 'end' specify the range over which to fit, and may be set to
    None to perform automatic ranging.

    'dX', 'amplitude', and 'dY' provide initial values for the
    corresponding parameters used in the nonlinear least squares fitting.  The
    function used for this fitting looks something like::

        # from xdp.interpolate.linear import interpolate
        y = interpolate(x + dX, x2, y)
        return (y * amplitude) + dY

    This functions returns a tuple of the form '(params, error)', where
    'params' is a tuple containing the best fit values for the dX, amplitude,
    and dY and 'error' is the value of chi-squared for this fit.
    """
    Min = max([ min(base.X), min(other.X) ])
    Max = min([ max(base.X), max(other.X) ])

    if start is None:
        start = Min
    if end is None:
        end = Min + 0.05 * (Max - Min)

    try:
        checkRange(Min, Max, start, end)
    except ValueError, e:
        raise MeansumError, ('energy shift %s' % e)

    e = end - abs(dX) * 2
    s = start + abs(dX) * 2

    x1, y1 = trim(base.X, base.Y, s, e)
    x2, y2 = trim(other.X, other.Y, start, end)

    func = IVShiftFunc(x2, y2)
    em = coljoin(x1, y1)
    params = [dX, amplitude, dY]
    return Scientific.Functions.LeastSquares.leastSquaresFit(func, params, em)


#
# Programs which use findIVShift() should still run without Scientific Python
# installed, but will generate a warning for every call to findIVShift()
#


try:
    import Scientific.Functions.LeastSquares
    import Scientific.Functions.FirstDerivatives
except ImportError:
    import warnings

    def findIVShift(base, other, start, end, dX, amplitude, dY):
        """I.V.S. correction is disabled - Scientific Python is not installed"""
        warnings.warn(findIVShift.__doc__, RuntimeWarning, 2)
        return [0.0, 0.0, 0.0], None
