# Name: xdp/tkutil.py
# Purpose: Useful Tkinter functions and classes
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""xdp.tkutil

Useful Tkinter and classes
"""


import os
import tkFont
import string
import os.path
import FileDialog
import tkFileDialog
from Tkinter import *


class StatusBar(Frame):
    """A simple application status bar.
    
    This class is from "An Introduction to Tkinter", where it was presented
    in tkSimpleStatusBar.py

    Copyright (c) 1997 by Fredrik Lundh

    fredrik@pythonware.com
    http://www.pythonware.com

    Docstrings and misc. hacking by Ken McIvor <mcivken@iit.edu>
    """

    def __init__(self, master=None, cnf={}, **kw):
        """Creates a new status bar widget.

        'master' is the Tkinter parent widget.
        """
        Frame.__init__(self, master, cnf, **kw)
        self.label = Label(self, bd=1, relief=SUNKEN, anchor=W, width=15,
            font='Helvetica 10')
        self.label.pack(fill=X)

    def get(self):
        """Returns the current contents of the status bar.
        """
        return self.label.cget('text')

    def set(self, format, *args):
        """Sets status bar's contents to be 'format % args'.
        """
        self.label.config(text=format % args)
        self.label.update_idletasks()

    def clear(self):
        """Clears the status bar's contents.
        """
        self.label.config(text="")
        self.label.update_idletasks()


class FrameLayoutMixin:
    """Tansparently puts its child class into a container Frame.

    This mixin is used to create compound widget classes which allow the
    programmer to access the main interior widget directly.  It is used by a
    lot of the other widgets in this module either directly or indirectly.  The
    'FileEntry' serves as a pretty clear example of how this mixin can be used.

    If this mixin is used with a class which is subclassed from the 'Tk' or
    'Toplevel' widget classes a 'RuntimeError' exception is raised.
    """
    def __init__(self, frame):
        """Creates a new frame layout container and overrides the appropriate
        geometry management methods.  This constructor should be invoked last,
        after the container frame has been created and  the widget's main
        constructor have been invoked.
        """
        func = lambda d, c=self.__class__: issubclass(c, d)
        if filter(func, (Tk, Toplevel)):
            raise RuntimeError, 'Cannot use FrameLayoutMixin with Tk or ' \
                'Toplevel widgets'
        if filter(func, (Canvas, Listbox, Text)):
            for method in Grid.__dict__.keys():
                if (method[0] != '_'
                  and method not in ('bbox', 'config', 'configure')):
                    setattr(self, method, getattr(frame, method))
        elif issubclass(self.__class__, Frame):
            for method in Grid.__dict__.keys():
                if method in ('grid', 'grid_forget', 'grid_remove'):
                    setattr(self, method, getattr(frame, method))
        else:
            for method in Grid.__dict__.keys():
                if method[0] != '_' and method not in ('config', 'configure'):
                    setattr(self, method, getattr(frame, method))

        for method in Pack.__dict__.keys():
            if method[0] != '_' and method not in ('config', 'configure'):
                setattr(self, method, getattr(frame, method))

        for method in Place.__dict__.keys():
            if method[0] != '_' and method not in ('config', 'configure'):
                setattr(self, method, getattr(frame, method))


class ScrolledMixin(FrameLayoutMixin):
    """This mixin can be used with widgets which support Tk's scrolling API
    (xscollcommand, xview, etc) to create a compound widget with scrollbars
    which automagically appear and disappear.  See the 'ScrolledText' class for
    a simple example, or the 'ScrolledCanvas' class for an example of how to
    accomplish the same thing when the child widget doesn't correctly update
    its scrollregion.
    """
    def __init__(self, frame):
        """Creates the container frame and scrollbars and attaches them to the
        widget.  This constructor should be invoked last, after the container
        frame has been created and  the widget's main constructor have been
        invoked.
        """
        frame.rowconfigure(0, weight=1)
        frame.rowconfigure(1, weight=0)
        frame.columnconfigure(0, weight=1)

        self.__showX = 0
        self.__showY = 0
        self.__update = 0
        self.__updateX = 0
        self.__updateY = 0

        self.__ysb = Scrollbar(frame, orient=VERTICAL)
        self.__xsb = Scrollbar(frame, orient=HORIZONTAL)

        relief = self.cget('relief')
        self.__xsb.configure(relief=relief)
        self.__ysb.configure(relief=relief)

        self.__grid = self.grid
        self.__grid_configure = self.grid_configure
        self.grid(row=0, column=0, rowspan=2, columnspan=2, sticky=NSEW)

        self.__xsb.grid(row=1, column=0, sticky='ewn')
        self.__ysb.grid(row=0, column=1, rowspan=2, sticky='nsw')

        self.__ysb.configure(command=self.yview)
        self.configure(yscrollcommand=self.__scrollY)

        self.__xsb.configure(command=self.xview)
        self.configure(xscrollcommand=self.__scrollX)

        apply(FrameLayoutMixin.__init__, (self, frame))
        Misc.tkraise(self)
        # Was the following, which breaks for Canvas widgets
#       self.tkraise()
        frame.update()
        frame.grid_propagate(0)

    def __scrollX(self, min, max):
        """Private method used as the callback for the widget's xscrollcommand.
        """
        self.__xsb.set(min, max)
        if abs(float(min) - float(max)) < 1:
            self.__updateX = 1
        else:
            self.__updateX = -1
        if not self.__update:
            self.__update = 1
            self.after(100, self.__updateScroll)

    def __scrollY(self, min, max):
        """Private method used as the callback for the widget's yscrollcommand.
        """
        self.__ysb.set(min, max)
        if abs(float(min) - float(max)) < 1 and not (min == '0' and max == '0'):
            self.__updateY = 1
        else:
            self.__updateY = -1
        if not self.__update:
            self.__update = 1
            self.after(100, self.__updateScroll)

    def __updateScroll(self):
        """Private method which draws or hides the scollbars as necessary,
        resizing the scrolled widget to fill as much of the frame as possible.
        """
        if self.__updateX == 1:
            self.__updateX = 0
            if not self.__showX:
                    yv = self.yview()
                    self.__grid_configure(rowspan=1)
                    self.__showX = 1
        elif self.__updateX == -1:
            self.__updateX = 0
            if self.__showX:
                self.__grid_configure(rowspan=2)
                self.__xsb.grid_configure(columnspan=1)
                self.__showX = 0

        if self.__updateY == 1:
            self.__updateY = 0
            if not self.__showY:
                    self.__grid_configure(columnspan=1)
                    self.__showY = 1
        elif self.__updateY == -1:
            self.__updateY = 0
            if self.__showY:
                self.__grid_configure(columnspan=2)
                self.__showY = 0

        if self.__showX:
            if self.__showY:
                self.__xsb.grid_configure(columnspan=1)
            else:
                self.__xsb.grid_configure(columnspan=2)

        self.__update = 0
        Misc.tkraise(self)


class ScrolledText(Text, ScrolledMixin):
    """A Tkinter 'Text' widget with scroll bars which appear as necessary.

    *Note*: there is currently a bug which will result in the horizontal
    scrollbar appearing and disappearing continuously under certain conditions
    (very long, unwrappable lines at the top or bottom of the visible area).
    """
    def __init__(self, root, **kw):
        """Creates a new 'ScrolledText' widget.

        'root' is the Tkinter parent widget.  Any keyword arguments are passed
        to the 'Text' widget's constructor.
        """
        self.__frame = Frame(root)
        apply(Text.__init__, (self, self.__frame,), kw)
        apply(ScrolledMixin.__init__, (self, self.__frame))


class ScrolledList(Listbox, ScrolledMixin):
    """A Tkinter 'Listbox' widget with scroll bars which appear as necessary.
    """
    def __init__(self, root, **kw):
        """Creates a new 'ScrolledList' widget.

        'root' is the Tkinter parent widget.  Any keyword arguments are passed
        to the 'Listbox' widget's constructor.
        """
        self.__frame = Frame(root)
        apply(Listbox.__init__, (self, self.__frame,), kw)
        apply(ScrolledMixin.__init__, (self, self.__frame))


class ScrolledCanvas(Canvas, ScrolledMixin):
    """A Tkinter 'Canvas' widget with scroll bars which appear as necessary.
    """
    def __init__(self, root, **kw):
        """Creates a new 'ScrolledCanvas' widget.

        'root' is the Tkinter parent widget.  Any keyword arguments are passed
        to the 'Canvas' widget's constructor.
        """
        self.__frame = Frame(root)
        apply(Canvas.__init__, (self, self.__frame,), kw)
        apply(ScrolledMixin.__init__, (self, self.__frame))

    def __scroll(self, event=None):
        """Private method to update the canvas's 'scrollregion' attribute.
        """
        coords = self.bbox(ALL)
        if coords:
            self.configure(scrollregion=(0, 0, coords[2], coords[3]))

    def create_arc(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_arc, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_bitmap(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_bitmap, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_image(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_image, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_line(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_line, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_oval(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_oval, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_polygon(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_polygon, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_rectangle(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_rectangle, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_text(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_text, (self,)+args, kw)
        self.__scroll()
        return rval

    def create_window(self, *args, **kw):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.create_window, (self,)+args, kw)
        self.__scroll()
        return rval

    def dchars(self, *args):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.dchars, (self,)+args)
        self.__scroll()
        return rval

    def delete(self, *args):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.delete, (self,)+args)
        self.__scroll()
        return rval

    def insert(self, *args):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.insert, (self,)+args)
        self.__scroll()
        return rval

    def move(self, *args):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.move, (self,)+args)
        self.__scroll()
        return rval

    def scale(self, *args):
        """Invokes the 'Canvas' widget method of the same name with the
        supplied arguments and keyword arguments, then updates the canvas's
        'scrollregion' attribute.
        """
        rval = apply(Canvas.scale, (self,)+args)
        self.__scroll()
        return rval


class ScrolledFrame(Frame, FrameLayoutMixin):
    """A Tkinter 'Frame' widget with scroll bars which appear as necessary.
    """
    def __init__(self, root, **kw):
        """Creates a new 'ScrolledFrame' widget.

        'root' is the Tkinter parent widget.  Any keyword arguments are passed
        to the 'Frame' widget's constructor.
        """
        self.__canvas = ScrolledCanvas(root)
        apply(Frame.__init__, (self, self.__canvas,), kw)
        apply(FrameLayoutMixin.__init__, (self, self.__canvas))

        self.bind('<Configure>', self.__configure)
        self.__canvas.create_window(0, 0, window=self, anchor=NW)

    def __configure(self, event):
        size = self.winfo_reqwidth(), self.winfo_reqheight()
        self.__canvas.configure(scrollregion="0 0 %s %s" % size)


class MessageText(ScrolledText):
    """A scrolled textbox with support for color-coded program output.

    Any output printed by the methods below will be displayed at the end of the
    message display, which will automatically scroll to bring it into view.

    This message display uses a ten point courier new font.
    """
    def __init__(self, root, readonly=1, **kw):
        """Creates a new 'MessageText' widget.

        'root' is the Tkinter parent widget.  If 'readonly' is true, then the
        widget will not accept user input.  
        """
        apply(ScrolledText.__init__, (self, root), kw)
#       cn = tkFont.Font()
#       cn.configure(family='courier new', size=10)
#       self.tag_config('alert', font=cn, foreground='blue')
#       self.tag_config('error', font=cn, foreground='red')
#       self.tag_config('normal', font=cn, foreground='black')
        self.tag_config('alert', foreground='blue')
        self.tag_config('error', foreground='red')
        self.tag_config('normal', foreground='black')

        if readonly:
            self.configure(state=DISABLED)
        self.__readonly = readonly

    def _write(self, text, tag):
        """Private method to output 'text' using the supplied list of tags
        'tag'.
        """
        if self.__readonly:
            self.configure(state=NORMAL)
        self.insert(END, text, (tag,))
        self.see(END)
        if self.__readonly:
            self.configure(state=DISABLED)

    def alert(self, text):
        """Prints 'text', followed by a newline, in blue text.
        """
        self._write('%s\n' % text, 'alert')

    def clear(self, event=None):
        """Clears the currently displayed output.
        """
        if self.__readonly:
            self.configure(state=NORMAL)
        self.delete('0.0', END)
        if self.__readonly:
            self.configure(state=DISABLED)

    def error(self, text):
        """Prints 'text', followed by a newline, in red text.
        """
        self._write('%s\n' % text, 'error')

    def normal(self, text):
        """Prints 'text', followed by a newline, in black text.
        """
        self._write('%s\n' % text, 'normal')

    def write(self, text):
        """Prints 'text' in black text.
        """
        self._write(text, 'normal')


class MultiFileDialog(FileDialog.LoadFileDialog):
    """A file dialog which allows users to select multiple files at once.

    Unlike a regular 'LoadFileDialog', this dialog's 'go()' method returns the
    list of selected files as a tuple.

    See the 'FileDialog' module for more information.
    """

    title = 'Open File(s)'

    def __init__(self, master, title=None):
        """Creates a new 'MultiFileDialog' dialog box.

        'master' is the Tkinter parent widget.  'title' is the optional window
        title, which defaults to "Open Files(s)".
        """
        apply(FileDialog.LoadFileDialog.__init__, (self, master, title))
        self.files.config(selectmode=EXTENDED)

    def ok_command(self):
        """'LoadFileDialog' method overridden to provide support for the
        selection of multiple files.
        """
        files = map(int, self.files.curselection())
        if not files:
            self.master.bell()
        else:
            files = map(self.files.get, files)
            files = map(lambda f, s=self: os.path.join(s.directory, f), files)
            for file in files:
                if not os.path.isfile(file):
                    self.master.bell()
                    return
            self.quit(files)


class FileSelector(ScrolledList, FrameLayoutMixin):
    """A scrolled listbox containing the names of selected files.
    
    Buttons allowing the user to add and remove files appear below the
    listbox.

    *Note*: the list of selected files should not be manipulated using the
    'Listbox' methods, but rather those specially provided by this class.
    """
    def __init__(self, root, files=[], **kw):
        """Create a new 'FileSelector' widget.

        'root' is the Tkinter parent widget.  'files' is an optional list of
        initial files to place in the listbox.
        """
        frame = Frame(root)
        apply(ScrolledList.__init__, (self, frame), kw)

        self.__files = files
        self.__directory = os.getcwd()
        self.__filter = '*'

        frame.columnconfigure(0, weight=1)
        frame.rowconfigure(0, weight=1)
        frame.rowconfigure(1, weight=0)
        self.grid(row=0, column=0, sticky=NSEW)

        bf = Frame(frame)
        Button(bf, text='Add...', command=self._add).grid(row=0, column=0,
            sticky=EW)
        Button(bf, text='Remove', command=self._remove).grid(row=0, column=1,
            sticky=EW)
        bf.grid(row=1, column=0, sticky='ewn')

        apply(FrameLayoutMixin.__init__, (self, frame))
        if self.__files:
            self._redraw()
        frame.update()

    def _add(self, event=None):
        """Creates a 'MultiFileDialog' to allow user to add files to the list.

        This function is the callback for the 'Add...' button.
        """
        dlg = MultiFileDialog(self)
        filter = StringVar()
        dlg.filter.configure(textvariable=filter)
        files = dlg.go(self.__directory, self.__filter)
        if files:
            self.__filter = filter.get()
            self.__files = self.__files + files
            self._redraw()

    def _remove(self, event=None):
        """Removes the selected file names from the list.

        This function is the callback for the 'Remove' button.
        """
        idx = map(int, self.curselection())
        if not idx:
            self.bell()
        else:
            rmCount = 0
            for i in idx:
                self.__files.pop(i - rmCount)
                rmCount = rmCount + 1
            self._redraw()

    def _redraw(self):
        """Private method used to redraw the list when it changes.

        This is necessary because the list actually displays the filenames
        after their common prefix, instead of their whole paths.  So,
        '/home/kmcivor/1.txt' and '/home/kmcivor/xray/TODO' would be displayed
        as '1.txt' and 'xray/TODO' respectively.
        """
#       self.__files.sort()
        self.delete(0, END)
        files = self.__expand(self.__files)
        base = os.path.commonprefix(files)
        dir = os.path.dirname(base)
        if len(base) > len(dir) + 1:
            if dir:
                base = dir + os.sep
            else:
                base = ''
        self.__directory = base
        f = map(lambda x, l=len(base): x[l:], files)
        map(lambda x, self=self: self.insert(END, x), f)
        self.update_idletasks()

    def __expand(self, files):
        """Private method used to normalize a list of files.
        """
        return map(os.path.abspath, map(os.path.expanduser, files))

    def getValue(self):
        """The current contents of the file selection list is returned as a list.
        """
        return list(self.__files)

    def setValue(self, files):
        """The contents of the file selection list is set to be the files named
        in the sequence 'files'.
        """
        self.__files = self.__expand(files)
        self.__redraw()


class LabelledFrame(Frame, FrameLayoutMixin):
    """An outlined frame with a label in the upper left hand corner.
    """
    def __init__(self, master, text, **kw):
        """Create a new 'LabelledFrame' widget.

        'root' is the Tkinter parent widget.  'text' is the text to be displayed
        in the label.
        """
        frame = Frame(master)
        self.__label = Label(frame, text=text)
        self.__border = Frame(frame, relief=GROOVE, bd=3)
        apply(Frame.__init__, (self, self.__border), kw)

        frame.rowconfigure(0, weight=1)
        frame.columnconfigure(0, weight=1)
        self.__border.rowconfigure(0, weight=1)
        self.__border.columnconfigure(0, weight=1)

        self.__label.grid(row=0, column=0, sticky=NW, padx=15, pady=1)
        self.__border.grid(row=0, column=0, sticky=NSEW, padx=10, pady=10)
        self.grid(row=0, column=0, sticky=NSEW, padx=7, pady=7)

        apply(FrameLayoutMixin.__init__, (self, frame))
        self.bind('<Configure>', self.__draw)
        frame.update()

    def __getitem__(self, key):
        """Special method allowing the user to retrieve the 'text' attribute in
        Tkinter's style.
        """
        if key == 'text':
            return self.__label.cget('text')
        else:
            return Widget.__getitem__(self, key)

    def cget(self, key):
        """Overrides the default 'cget' method to allow the user to retrieve
        the 'text' attribute.
        """
        return self.__getitem__(key)

    def __setitem__(self, key, value):
        """Special method allowing the user to set the 'text' attribute in
        Tkinter's style.
        """
        if key == 'text':
            self.__label.configure(text=value)
            self.__draw()
        else:
            return Widget.configure(self, {key:value})

    def configure(self, cnf=None, **kw):
        """Overrides the default 'configure' method to allow the user to set
        the 'text' attribute.
        """
        if cnf and cnf.has_key('text'):
            self.__setitem__('text', cnf['text'])
            del cnf['text']
        if kw.has_key('text'):
            self.__setitem__('text', kw['text'])
            del kw['text']
        apply(Widget.configure, (self, cnf), kw)

    def __draw(self, event=None):
        """Private method which redraws the label.

        This function serves as a callback for the '<Configure>' event.
        """
        self.__label.tkraise()


class AutoRangeEntryElement(Frame):
    """A compound widget containing two radio buttons, the second of which has
    an entry field associated with it.
    """
    __inf = float('inf')

    def __init__(self, root, text1='', text2='', default=None, conv=float, **kw):
        """Create a new 'AutoRangeEntryElement' widget.

        'root' is the Tkinter parent widget.  'text1' is the label of the first
        radio button and 'text2' is the label of the second radio button.
        'default' is the default value for the entry field.  If it is
        specified, then the second radio button is selected.  Otherwise, the
        first radio button is.  'conv' is the conversion function used to
        transform the text of the entry box into some other value.
        """
        apply(Frame.__init__, (self, root), kw)

        self.__conv = conv
        self.__specify = IntVar()

        self.__auto = Radiobutton(self, text=text1, variable=self.__specify,
            value=0)
        self.__spec = Radiobutton(self, text=text2, variable=self.__specify,
            value=1)

        self.__text = Entry(self, width=10)
        if default is None:
            self.__specify.set(0)
        else:
            self.__specify.set(1)
            self.__text.insert(END, str(default))

        self.rowconfigure(1, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(0, weight=0)
        self.__auto.grid(row=0, column=0, sticky=W)
        self.__spec.grid(row=1, column=0, sticky=W)
        self.__text.grid(row=1, column=1, stick=EW)

    def getValue(self):
        """Returns the value of this input widget.
        
        If the first radio button is selected then 'None' is returned.
        Otherwise the text in the entry field is convered to some other value
        using the 'conv' function specified in the constructor and returned.
        """
        if not self.__specify.get():
            return None
        else:
            rval = self.__conv(self.__text.get())
            if rval == self.__inf or rval == -1 * self.__inf:
                raise ValueError, 'floating point overflow'
            else:
                return rval

class AutoRangeEntry(Frame):
    """Compound widget which allows the user to enter the start and and points of
    some range.
    """
    def __init__(self, root, start=None, end=None, name=None, orient=VERTICAL,
        conv=float, **kw):
        """Create a new 'AutoRangeEntry' widget.

        'root' is the Tkinter parent widget.  'start' and 'end' are the values
        to display in the start and end points' entry boxes.  If one of them is
        specified, then the radio button for that entry box is selected,
        otherwise the "auto" one is.  'name' is the descriptive name of this
        range.  This information is used when generating exceptions in
        'getValue()'.  'orient' controls how the 'AutoRangeEntryElement' frames
        are layed out on the screen and is one of 'VERTICAL' or 'HORIZONTAL'.
        'conv' is the conversion function supplied to the
        'AutoRangeEntryElement' frames.  Any keyword arguments are passed to
        the 'Frame' widget's constructor.
        """
        self.__name = name
        apply(Frame.__init__, (self, root), kw)
        self.__start = AutoRangeEntryElement(self, 'Auto', 'Specify:', start, conv)
        self.__end = AutoRangeEntryElement(self, 'Auto', 'Specify:', end, conv)

        if orient == VERTICAL:
            self.columnconfigure(0, weight=0)
            self.columnconfigure(1, weight=1)
            self.rowconfigure(0, weight=1)
            self.rowconfigure(1, weight=1)
            self.rowconfigure(2, weight=1)
        else:
            self.rowconfigure(0, weight=1)
            self.columnconfigure(0, weight=0)
            self.columnconfigure(1, weight=1)
            self.columnconfigure(2, weight=0)
            self.columnconfigure(3, weight=0)
            self.columnconfigure(4, weight=1)

        Label(self, text='Start:').grid(row=0, column=0, padx=5, sticky=NE)
        self.__start.grid(row=0, column=1, sticky='ewn')

        if orient == VERTICAL:
            Frame(self, height=10).grid(row=1, column=0, columnspan=2,
                sticky='ewn')
            Label(self, text='End:').grid(row=2, column=0, padx=5, sticky=NE)
            self.__end.grid(row=2, column=1, sticky='ewn')
        else:
            Frame(self, width=10).grid(row=0, column=2, sticky='ewn')
            Label(self, text='End:').grid(row=0, column=3, padx=5, sticky=NE)
            self.__end.grid(row=0, column=4, sticky='ewn')

    def getValue(self):
        """Returns the range the user has input.
        
        The start and end points are returned in a 2-tuple of the form '(start,
        end)'.  Either may be 'None', indicating the the first radio button
        ("Auto") was selected for that point.

        A 'ValueError' exception may be raised if the fields contain invalid
        values.
        """
        try:
            start = self.__start.getValue()
        except ValueError:
            if not self.__name:
                raise ValueError, 'invalid value for start of range'
            else:
                raise ValueError, ('invalid value for start of %s range'
                    % self.__name)
        try:
            end = self.__end.getValue()
        except ValueError:
            if not self.__name:
                raise ValueError, 'invalid value for end of range'
            else:
                raise ValueError, ('invalid value for end of %s range'
                    % self.__name)
        return (start, end)


class FileEntry(Entry, FrameLayoutMixin):
    """A labelled entry field for inputting a file name.

    The entry field has a "..." button to its right which opens a file dialog
    and allows the user to select a new value for the field.
    """
    def __init__(self, root, text='', open=0, directory=None, types=None,
        extension=None, **kw):
        """Create a new 'FileEntry' widget.

        'root' is the Tkinter parent widget.  'text' is the label associated
        with this input.  'open' is a boolean flag indicating whether this file
        is to be opened or saved to.  'directory' is the directory to initially
        open the file dialog in.  'types' and 'extension' are passed to the
        tkFileDialog functions.  Any keyword arguments are passed to the
        'Frame' widget's constructor.
        """

        frame = Frame(root)
        apply(Entry.__init__, (self, frame), kw)

        self.__open = open
        self.__types = types or (('All Files', '*.*'),)
        self.__extension = extension or '*.*'
        self.__directory = directory or os.getcwd()
        self.__button = Button(frame, text='...', command=self._open,
            width=3, padx=0, pady=0)

        if text:
            frame.columnconfigure(1, weight=1)
            self.__label = Label(frame, text=text, anchor=E)
            self.__label.grid(row=0, column=0, sticky=E)
            self.grid(row=0, column=1, sticky='ew', padx=5)
            self.__button.grid(row=0, column=2, sticky=W)
        else:
            frame.columnconfigure(0, weight=1)
            self.grid(row=0, column=0, sticky='ew', padx=5)
            self.__button.grid(row=0, column=1, sticky=W)

        apply(FrameLayoutMixin.__init__, (self, frame))

    def _open(self, event=None):
        """Opens the appropriate tkFileDialog and gets the user's new file
        selection.  If they make one it will replace the current contents of
        the entry field.
        """
        if self.__open:
            fname = tkFileDialog.askopenfilename(initialdir=self.__directory,
                    defaultextension=self.__extension,
                    filetypes=self.__types,
                    title='Open File',
                    parent=self
                    )
        else:
            fname = tkFileDialog.asksaveasfilename(initialdir=self.__directory,
                    defaultextension=self.__extension,
                    filetypes=self.__types,
                    title='Save Data',
                    parent=self
                    )
        if not fname:
            return

        path, ext = os.path.splitext(os.path.expanduser(fname))
        self.__extension = string.lower(ext) or self.__extension
        self.__directory = os.path.split(path)[0]
        self.delete(0, END)
        self.insert(END, fname)


#
# Test/Demo Code
#


def _testScrolledText(win):
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.title('ScrolledText Demo')
    st = ScrolledText(win, wrap=WORD, width=40, height=12)
    st.grid(row=0, column=0, sticky=NSEW)
    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testScrolledList(win):
    import string
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.title('ScrolledList Demo')
    sl = ScrolledList(win)
    sl.grid(row=0, column=0, sticky=NSEW)
    entries = map(str, range(0, 15))
    for i in entries:
        sl.insert(END, i)
    sl.insert(END, string.join(entries))
    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testScrolledCanvas(win):
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.title('ScrolledCanvas Demo')
    sc = ScrolledCanvas(win)
    sc.grid(row=0, column=0, sticky=NSEW)
    sc.create_rectangle('5c','5c','100c','10c', outline='black')
    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testFileSelector(win):
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.rowconfigure(0, weight=1)
    win.rowconfigure(1, weight=0)
    win.title('FileSelector Demo')
    fs = FileSelector(win)
    fs.grid(row=0, column=0, sticky=NSEW)

    def pFiles(fs):
        print 'Files:'
        for file in fs.getValue():
            print file
        print

    cmd = lambda event=None, fs=fs, pFiles=pFiles: pFiles(fs)

    Button(win, text='Print File List', command=cmd).grid(row=1, column=0,
        sticky='ewn')

    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testLabelledFrame(win):
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.title('LabelledFrame Demo')
    lf = LabelledFrame(win, text='LabelText')
    lf.rowconfigure(0, weight=1)
    lf.columnconfigure(0, weight=1)
    sl = ScrolledList(lf)
    sl.grid(row=0, column=0, sticky=NSEW)
    lf.grid(row=0, column=0, sticky=NSEW)
    entries = map(str, range(0, 15))
    for i in entries:
        sl.insert(END, i)
    sl.insert(END, string.join(entries))
    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testAutoRangeEntryElement(win):
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.title('AutoRangeEntryElement Demo')
    are = AutoRangeEntryElement(win, default=3.1415)
    are.grid(row=0, column=0, sticky=NSEW)

    def pValue(are):
        try:
            print str(are.getValue())
        except ValueError, e:
            print 'Error:', e

    cmd = lambda event=None, are=are, pValue=pValue: pValue(are)

    Button(win, text='Print Current Value', command=cmd).grid(row=1, column=0,
        sticky='ewn')

    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testAutoRangeEntry_Vertical(win):
    win.columnconfigure(0, weight=1)
    win.rowconfigure(0, weight=1)
    win.rowconfigure(1, weight=0)
    win.title('AutoRangeEntry Demo')
    are = AutoRangeEntry(win, end=3.1415, name='test')
    are.grid(row=0, column=0, sticky='ewn')

    def pValue(are):
        try:
            print str(are.getValue())
        except ValueError, e:
            print 'Error:', e

    cmd = lambda event=None, are=are, pValue=pValue: pValue(are)

    Button(win, text='Print Current Value', command=cmd).grid(row=1, column=0,
        sticky='ewn')

    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testAutoRangeEntry_Horizontal(win):
    win.columnconfigure(0, weight=1)
    win.rowconfigure(0, weight=1)
    win.rowconfigure(1, weight=0)
    win.title('AutoRangeEntry Demo')
    are = AutoRangeEntry(win, end=3.1415, name='test', orient=HORIZONTAL)
    are.grid(row=0, column=0, sticky='ewn')

    def pValue(are):
        try:
            print str(are.getValue())
        except ValueError, e:
            print 'Error:', e

    cmd = lambda event=None, are=are, pValue=pValue: pValue(are)

    Button(win, text='Print Current Value', command=cmd).grid(row=1, column=0,
        sticky='ewn')

    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _testFileEntry(win):
    win.rowconfigure(0, weight=1)
    win.columnconfigure(0, weight=1)
    win.rowconfigure(0, weight=1)
    win.rowconfigure(1, weight=0)
    win.title('FileEntry Demo')
    fe = FileEntry(win, text='File:')
    fe.grid(row=0, column=0, sticky=NSEW)

    def pFile(fs):
        print 'File:', fs.get()

    cmd = lambda event=None, fe=fe, pFile=pFile: pFile(fe)

    Button(win, text='Print File', command=cmd).grid(row=1, column=0,
        sticky='ewn')

    win.wm_protocol('WM_DELETE_WINDOW', lambda w=win,*a: (w.destroy(), w.quit()))
    win.mainloop()


def _runTests():
    import sys
    root = Tk()
    root.title('tkutil demos')
    g = globals()
    funcNames = filter(lambda x: x[:5]=='_test', g.keys())
    funcNames.sort()
    for func in funcNames:
        f = lambda tk=root, f=g[func]: f(Toplevel(tk))
        Button(root, text=func[5:], width=20, command=f).grid(sticky=NSEW)
    Button(root, text='Exit', width=20, command=sys.exit).grid(sticky=NSEW)
    root.columnconfigure(0, weight=1)
    for i in range(0, root.grid_size()[1]):
        root.rowconfigure(i, weight=1)
    root.mainloop()


if __name__ == '__main__':
    _runTests()
