## Automatically adapted for numpy.oldnumeric Feb 04, 2011 by 

# Name: xdp/transformation.py
# Purpose: XDP object-oriented data manipulation
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""xdp.transformation

This module contains XDP's mappings and transformations, which provide an
object-oriented API for manipulating data.
"""


import numpy.oldnumeric as Numeric

import xdp as _xdp
import data as _data
import interpolate as _interpolate


class Mapping:
    """
    Mapping objects are used to remap the column names of xdp.Data objects.

    They allow user-specific column names to be abstracted from
    application-specific ones and subsequently simplify software which must
    support arbitrary column names.

    Mapping objects are applied to xdp.Data objects using Python's
    function-call operator.  In other words, call the mapping object as though
    it were a function which takes one argument, an xdp.Data object.

    Example:
        RemapXafsTransmission = Mapping(Energy='Energy', Io='mcs3', It='mcs4')
        transData = RemapXafsTransmission(userData)

    Mapping may be inherited in order to provide application-specific mappings.
    When an inherited mapping is applied to an xdp.Data object, any methods
    the child class define which have names in the form of 'make_<COLUMN>' will
    be be called to generate the vector for the column '<COLUMN>'.  Such
    methods may return None, in order to supress the addition of the column.
    See XafsTransmissionMapping for an example.
    """

    def __init__(self, **kwds):
        """Create a new mapping from the keyword arguments.

        The keyword arguments contain all of the mapping this object will
        perform, in the form destination=source.
        """
        self.__map = kwds

    def finish(self, data):
        """Virtual method which is called to finalize the xdp.Data object
        'data' before it is returned.
        """
        pass

    def __call__(self, data):
        """Apply this mapping to the xdp.Data object 'data'.

        An xdp.ColumnNameError exception is raised if any of the source column
        names are invalid.
        """
        result = data.__class__()
        for attrName, columnName in self.__map.items():
            if columnName is None:
                continue
            column = data.get(columnName)
            result.add(attrName, column)

        for callbackName in dir(self) + dir(self.__class__):
            if not callbackName.startswith('make_'):
                continue

            callback = getattr(self, callbackName)
            if not callable(callback):
                continue

            attrName = callbackName[5:]
            if attrName in result.columns():
                continue

            column = callback(data, result)
            if column is not None:
                result.add(attrName, column)

        self.finish(result)
        return result


class _ReferenceChannelImplementation:
    """FIXME: write docstring!
    """
    def __init__(self, Iref):
        """FIXME: write docstring!
        """
        self.__Iref = Iref

    def make_Iref(self, data, result):
        """If a reference channel expression was supplied when this mapping was
        created, use it to calculate Iref.
        """
        if self.__Iref is None:
            return None

        try:
             Iref = data.evaluate(self.__Iref)
        except _xdp.EvaluationError, e:
            raise _xdp.EvaluationError(
                '%s in reference channel expression "%s"'
                % (e, self.__Iref))

        if not isinstance(Iref, Numeric.ArrayType) or len(Iref.shape) != 1:
            raise xdp.EvaluationError(
                ('reference channel expression "%s" does not evaluate to a '
                + 'one-dimensional Numeric array') % self.__Iref)

        return Iref


class XafsMapping(Mapping):
    """XafsMapping objects encapsulate the mapping for XAFS experiments which
    contain both fluorescence and transmission data.

    The destination columns are:
        'Energy' -- energy values for each data point
        'If'     -- fluorescence detector intensity
        'Io'     -- initial beam intensity
        'It'     -- transmission detector intensity

    'Iref' is the optional reference channel expression.  If it is not None, it
    is evaluated via xdp.evaluate() to create the reference channel's column.
    """

    def __init__(self, Energy='energy', If='If', Io='Io', It='It', Iref=None):
        """Create a new XafsMapping object, assigning source column names via
        keyword arguments.
        """
        Mapping.__init__(self, Energy=Energy, If=If, Io=Io, It=It)
        self.__Iref = _ReferenceChannelImplementation(Iref)

    def make_Iref(self, data, result):
        return self.__Iref.make_Iref(data, result)


class XafsTransmissionMapping(Mapping):
    """XafsTransmissionMapping objects encapsulate the mapping for XAFS
    experiments which contain transmission data.

    The destination columns are:
        'Energy' -- energy values for each data point
        'Io'     -- initial beam intensity
        'It'     -- transmission detector intensity

    The destination column 'Mu' is calculated from 'It' and 'Io' using the
    formula 'Mu = log(Io/It)'.
    """

    def __init__(self, Energy='energy', It='It', Io='Io', Iref=None):
        """Create a new XafsTransmissionMapping object, assigning source column
        names via keyword arguments.
        """
        Mapping.__init__(self, Energy=Energy, It=It, Io=Io)
        self.__Iref = _ReferenceChannelImplementation(Iref)

    def make_Mu(self, data, result):
        """Calculate Mu when the mapping object is applied to 'data'.
        """
        return Numeric.log(result.Io / result.It)

    def make_Iref(self, data, result):
        return self.__Iref.make_Iref(data, result)


class XafsFluorescenceMapping(Mapping):
    """XafsFluorescenceMapping objects encapsulate the mapping for XAFS
    experiments which contain fluorescence data.

    The destination columns are:
        'Energy' -- energy values for each data point
        'If'     -- fluorescence detector intensity
        'Io'     -- initial beam intensity

    The destination column 'Mu' is calculated from 'IF' and 'Io' using the
    formula 'Mu = If/Io'.
    """

    def __init__(self, Energy='energy', If='If', Io='Io', Iref=None):
        """Create a new XafsFluorescenceMapping object, assigning source column
        names via keyword arguments.
        """
        Mapping.__init__(self, Energy=Energy, If=If, Io=Io)
        self.__Iref = _ReferenceChannelImplementation(Iref)

    def make_Mu(self, data, result):
        """Calculate Mu when the mapping object is applied to 'data'.
        """
        return result.If / result.Io

    def make_Iref(self, data, result):
        return self.__Iref.make_Iref(data, result)


#
##
##  Transformations
##
#

class Transformation:
    """Transformation objects are used to apply arbitrary transformations to
    xdp.Data objects, resulting in new xdp.Data objects which reflect those
    changes.

    The Transformation class does not define a constructor or any actual
    transformation.  It is intended to act as the parent class for classes
    implementing real transformations.

    Any columns not modified by a transformation are not included in the
    resulting object.

    Transformation objects are applied to xdp.Data objects using Python's
    function-call operator.  In other words, call the transformation object as
    though it were a function which takes one argument, an xdp.Data object.
    """

    def __call__(self, data):
        """Apply this transformation to the xdp.Data object 'data'.
        """
        return self.apply(data)

    def apply(self, data):
        """Virtual method which implements the transformation and returns a new
        data object.
        """
        return data


class TrimTransform(Transformation):
    """TrimTransform objects encapsulate the process of triming a pair of
    vectors, X and Y, to some range of X values.
    """

    def __init__(self, xName, yName, start, end):
        """Create a new TrimTransfrom object which will trim the X vector from
        column 'xName' and the Y vetor from column 'yName' to X values from
        'start' to 'end', inclusive.
        """
        self.__xName = xName
        self.__yName = yName
        self.__start = start
        self.__end = end

    def apply(self, data):
        """Trim the X and Y columns from the object 'data' and return a new
        data object containing the trimmed columns.

        An xdp.ColumnNameError is raised if either of the source column names
        specified in the constructor are invalid.
        """
        x = data.get(self.__xName)
        y = data.get(self.__yName)

        x, y = _data.trim(x, y, self.__start, self.__end)
        result = data.__class__()
        result.set(self.__xName, x)
        result.set(self.__yName, y)

        return result


class CleanTransform(Transformation):
    """CleanTransform objects encapsulate the process of cleaning a pair of
    vectors, X and Y, by removing (X,Y) pairs which have duplicate X values.
    """

    def __init__(self, cName, *columnNames):
        """Create a new CleanTransfrom object which will clean the vector
        from column 'cName' and any additional vectors from 'columnNames'.
        """
        self.__cName = cName
        self.__columnNames = columnNames

    def apply(self, data):
        """Clean the vectors columns from the object 'data' and return a new
        data object containing the trimmed columns.

        An xdp.ColumnNameError is raised if either of the source column names
        specified in the constructor are invalid.
        """
        v = data.get(self.__cName)
        vectors = [ data.get(x) for x in self.__columnNames ]

        cleaned = _data.clean(v, *vectors)
        result = data.__class__()
        result.set(self.__cName, cleaned[0])
        for i in range(0, len(self.__columnNames)):
            result.set(self.__columnNames[i], cleaned[i+1])

        return result


class LintTransform(Transformation):
    """LintTransform objects encapsulate the process of using linear
    interpolation to interplolate a pair of vectors, X and Y, to some abscissa.
    """

    def __init__(self, xName, yName, grid):
        """Create a new LintTransfrom object which will interpolate the X
        vector from column 'xName' and the Y vetor from column 'yName' to the
        abscissa 'grid', a one-dimensional Numeric array.
        """
        self.__xName = xName
        self.__yName = yName
        self.__grid = grid

    def apply(self, data):
        """Interpolate the X and Y columns from the object 'data' and return a
        new data object containing the abscissa in place of the X column and
        the intolpolated values in place of the Y column.

        An xdp.ColumnNameError is raised if either of the source column names
        specified in the constructor are invalid.
        """
        x = data.get(self.__xName)
        y = data.get(self.__yName)

        y = _interpolate.lint(x, y, self.__grid)
        result = data.__class__()
        result.set(self.__xName, self.__grid)
        result.set(self.__yName, y)

        return result


# A complete example - regrid some files
#
#def regridFiles(fileNames):
#   transRemap = XafsTransmissionMapping(Energy='energy', Io='mcs3', It='mcs4')
#
#   gridHeader, gridData = xdp.io.readFile(fileNames[0])
#   gridData = transRemap(gridData)
#
#   regrid = LintTransformation('Energy', 'Mu', gridData.Energy)
#
#   for fileName in fileNames[1:]:
#       header, data = xdp.io.readFile(fileName)
#       data = regrid(transRemap(data))
#       xdp.io.writeFile('regrid-%s' % fileName, header, data)


class RebinTransform(Transformation):
    """RebinTransform objects encasulate the process of performing XAFS data
    reduction on one or more vectors.

    See xdp.data.rebin() for more information about rebinning.
    """

    def __init__(self, Eo, offset, estep, kstep, energyName, *columnNames):
        """Create a new RebinTransfrom object which will rebin the columns
        'columnNames' using the energy values from 'energyName'.

        'Eo', 'offset', 'estep', and 'kstep' are all parameters to
        xdp.data.rebin(), and mean the same thing here.
        """
        self.__Eo = Eo
        self.__offset = offset
        self.__estep = estep
        self.__kstep = kstep
        self.__energyName = energyName
        self.__columnNames = columnNames

    def apply(self, data):
        """Rebin the energy column EnergyName and columns of ColumnNames from
        the object 'data' and return a new Data object containing the rebinned
        values in place of the originals.

        An xdp.ColumnNameError is raised if any of the source column names
        specified in the constructor are invalid.
        """
        vectors = [ data.get(self.__energyName) ]
        for columnName in self.__columnNames:
            vectors.append(data.get(columnName))

        matrix = _data.coljoin(*vectors)
        rebinned = _data.rebin(matrix, 0, self.__Eo, self.__offset,
            self.__estep, self.__kstep)

        result = data.__class__()
        result.set(self.__energyName, rebinned[:,0])

        for i in range(0, len(self.__columnNames)):
            result.set(self.__columnNames[i], rebinned[:,i+1])

        return result
