## Automatically adapted for numpy.oldnumeric Feb 04, 2011 by 

# Name: xdp/xafs.py
# Purpose: utilities for writing XDP applications which process XAFS data
# Author: Ken McIvor <mcivken@iit.edu>
#
# Copyright 2003-2004 Illinois Institute of Technology
#
# See the file "LICENSE" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

"""A library of supporting classes and functions for writing XDP applications
which process XAFS data.
"""


import os
import numpy.oldnumeric as Numeric

import xdp
from xdp.io import readFile
from xdp.interpolate.linear import interpolate
from xdp.data import clean, checkRange, coljoin, rebin, RebinError, trim
from xdp.transformation import CleanTransform, LintTransform, TrimTransform, RebinTransform


class XafsData(xdp.Data):
    """XafsData objects represent XAFS experiment data from a file.

    Attributes:
        _path = full path and name of this data's source file
        _file = name of this data's source file
        _shift = relative energy shift of this data, or None
        _header = header for this data's source file, or None
    """
    def __init__(self, path='', mapping=None):
        """Create a new XafsData object, optionally loading the XAFS experiment
        data from the file 'path' and remapping it using 'mapping'.

        If the 'path' keyword argument is specified, the 'mapping' keyword
        argument must also be specified.
        """
        if path and not mapping:
            raise ValueError('"path" keyword arguments required "mapping"')

        xdp.Data.__init__(self)

        self._path = path
        self._shift = None
        self._header = None
        self._file = os.path.basename(path)

        if path:
            header, data = readFile(path)
            data = mapping(data)
            if data.has('Iref'):
                clean = CleanTransform('Energy', 'Mu', 'Iref')
            else:
                clean = CleanTransform('Energy', 'Mu')
            data = clean(data)

            self._header = header
            for columnName in data.columns():
                self.add(columnName, data.get(columnName))


def shift(xafsData, start, end):
    """Perform energy shift correction relative to the first XafsData object of
    the sequence 'xafsData', updating the elements of that sequence to reflect
    the energy shifts.

    The arguments 'start' and 'end' specify the energy range over which to
    perform the shift correction.
    """
    Min = max([ min(x.Energy) for x in xafsData ])
    Max = min([ max(x.Energy) for x in xafsData ])

    try:
        checkRange(Min, Max, start, end)
    except ValueError, e:
        raise ValueError('energy shift %s' % e)

    trim = TrimTransform('Energy', 'Mu', start, end)
    base = trim(xafsData[0])

    for xafs in xafsData[1:]:
        other = trim(xafs)
        params, error = findEnergyShift(base, other, None, None, 0.001, 1.0,
            0.001)

        xafs._shift = params[0]
        xafs.Energy += xafs._shift


def shiftRef(xafsData, start, end):
    """Perform energy shift correction using reference channels.

    The arguments 'start' and 'end' specify the energy range over which to
    perform the shift correction.
    """
    Min = max([ min(x.Energy) for x in xafsData ])
    Max = min([ max(x.Energy) for x in xafsData ])

    try:
        checkRange(Min, Max, start, end)
    except ValueError, e:
        raise ValueError('energy shift %s' % e)

    trim = TrimTransform('Energy', 'Iref', start, end)
    base = trim(xafsData[0])
    base.rename('Iref', 'Mu')

    for xafs in xafsData[1:]:
        other = trim(xafs)
        other.rename('Iref', 'Mu')

        params, error = findEnergyShift(base, other, None, None, 0.001, 1.0,
            0.001)

        xafs._shift = params[0]
        xafs.Energy += xafs._shift


def sum(xafsData):
    """Sum XAFS experiemnt data by regridding Mu for each XafsData object in
    the 'xafsData' sequence to the first element of the sequence using linear
    interpolation and then summing the results.

    The return value is an xdp.Data object containing the Energy column of the
    first XafsData object of 'xafsData' and the sum of the Mu columns of all of
    'xafsData'.
    """
    start = max([ min(x.Energy) for x in xafsData ])
    end = min([ max(x.Energy) for x in xafsData ])

    trim = TrimTransform('Energy', 'Mu', start, end)
    sum = trim(xafsData[0])

    interpolate = LintTransform('Energy', 'Mu', sum.Energy)

    for xafs in xafsData[1:]:
        xafs = interpolate(xafs)
        sum.Mu += xafs.Mu

    return sum


def mean(xafsData):
    """Find the arithmetic mean XAFS experiemnt data by summing 'xafsData'
    using sum() and then dividing the result's Mu by the number of elements.

    The return value is an xdp.Data object containing the Energy column of the
    first XafsData object of 'xafsData' and the arithmetic mean of the Mu
    columns of all of 'xafsData'.
    """
    result = sum(xafsData)
    result.Mu /= float(len(xafsData))
    return result


def rebin(xafsData, Eo, offset, estep, kstep):
    """Rebin the Energy and Mu columns of XafsData objects in the sequence
    'xafsData', updating the elements of that sequence to reflect the rebinned
    values.

    'Eo', 'offset', 'estep', and 'kstep' are all parameters for an
    xdp.transformation.RebinTransform, and mean the same thing here.
    """
    if xafsData[0].hasColumn('Iref'):
        columns = ('Mu', 'Iref')
    else:
        columns = ('Mu',)
    rebin = RebinTransform(Eo, offset, estep, kstep, 'Energy', *columns)

    for i in range(0, len(xafsData)):
        xafsData[i] = rebin(xafsData[i])


#
# Least-squares fitting... requires Scientific Python
#

class EnergyShiftFunc:
    """Private utility class implementing the callback function for
    findEnergyShift().
    """
    def __init__(self, energy, mu):
        """Initialize a new instance of the callback function.  'energy' and
        'mu' are the vectors of the second data set from a call to
        findEnergyShift().
        """
        self.__energy = energy
        self.__mu = mu

    def __call__(self, params, energy):
        """Implements the callback function for the nonlinear least squared fit.
        """
        dEnergy = params[0]
        amplitude = params[1]
        dMu = params[2]
        mu = interpolate(energy - dEnergy, self.__energy, self.__mu)
        return (mu * amplitude) + dMu


def findEnergyShift(base, other, start, end, dEnergy, amplitude, dMu):
    """Finds the delta-Energy, amplitude, and delta-Mu that fits two XAFS data
    sets, 'base' and 'other'.

    The xdp.Data object 'other' is fit the xdp.Data object 'base' using the
    columns 'Energy' and 'Mu'.

    'start' and 'end' specify the range over which to fit, and may be set to
    None to perform automatic ranging.

    'dEnergy', 'amplitude', and 'dMu' provide initial values for the
    corresponding parameters used in the nonlinear least squares fitting.  The
    function used for this fitting looks something like::

        # from xdp.interpolate.linear import interpolate
        mu = interpolate(energy - dMu, energy2, mu2)
        return (mu * amplitude) + dMu

    This functions returns a tuple of the form '(params, error)', where
    'params' is a tuple containing the best fit values for the delta-Energy,
    amplitude, and delta-Mu and 'error' is the value of chi-squared for this
    fit.
    """
    Min = max([ min(base.Energy), min(other.Energy) ])
    Max = min([ max(base.Energy), max(other.Energy) ])

    try:
        start, end = checkRange(Min, Max, start, end)
    except ValueError, e:
        raise RebinError, ('energy shift %s' % e)

    e = end - abs(dEnergy) * 2
    s = start + abs(dEnergy) * 2

    energy1, mu1 = trim(base.Energy, base.Mu, s, e)
    energy2, mu2 = trim(other.Energy, other.Mu, start, end)

    func = EnergyShiftFunc(energy2, mu2)
    em = coljoin(energy1, mu1)
    params = [dEnergy, amplitude, dMu]
    return Scientific.Functions.LeastSquares.leastSquaresFit(func, params, em)


#
# Programs which use findEnergyShift() should still run without Scientific
# Python installed, but it will generate a warning for every call to
# findEnergyShift()
#


try:
    import Scientific.Functions.LeastSquares
    import Scientific.Functions.FirstDerivatives
except ImportError:
    import warnings

    def findEnergyShift(base, other, start, end, dEnergy, amplitude, dMu):
        warnings.warn(
    'Energy shift correction is disabled - Scientific Python is not installed',
            RuntimeWarning, 2)

        return [0.0, 0.0, 0.0], None
